@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2016 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2017 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -12,18 +12,19 @@ from xpra.log import Logger
 log = Logger("gobject", "client")
 
 import sys
-from xpra.util import nonl, sorted_nicely, print_nested_dict, envbool, DONE
-from xpra.os_util import bytestostr, get_hex_uuid
+from xpra.util import nonl, sorted_nicely, print_nested_dict, envint, flatten_dict, DONE
+from xpra.os_util import bytestostr, get_hex_uuid, PYTHON3, POSIX, OSX
 from xpra.client.client_base import XpraClientBase, EXTRA_TIMEOUT
-from xpra.exit_codes import (EXIT_OK, EXIT_TIMEOUT, EXIT_INTERNAL_ERROR, EXIT_FAILURE, EXIT_UNSUPPORTED, EXIT_REMOTE_ERROR, EXIT_FILE_TOO_BIG)
+from xpra.exit_codes import (EXIT_OK, EXIT_CONNECTION_LOST, EXIT_TIMEOUT, EXIT_INTERNAL_ERROR, EXIT_FAILURE, EXIT_UNSUPPORTED, EXIT_REMOTE_ERROR, EXIT_FILE_TOO_BIG)
 
-FLATTEN_INFO = envbool("XPRA_FLATTEN_INFO", True)
+FLATTEN_INFO = envint("XPRA_FLATTEN_INFO", 1)
 
 
-class GObjectXpraClient(XpraClientBase, gobject.GObject):
+class GObjectXpraClient(gobject.GObject, XpraClientBase):
     """
         Utility superclass for GObject clients
     """
+    INSTALL_SIGNAL_HANDLERS = True
 
     def __init__(self):
         gobject.GObject.__init__(self)
@@ -31,7 +32,8 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
 
     def init(self, opts):
         XpraClientBase.init(self, opts)
-        self.install_signal_handlers()
+        if self.INSTALL_SIGNAL_HANDLERS:
+            self.install_signal_handlers()
         self.glib_init()
         self.gobject_init()
 
@@ -52,7 +54,7 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
         #overriden in subclasses!
         return "Python%s/GObject" % sys.version_info[0]
 
-    def timeout(self, *args):
+    def timeout(self, *_args):
         log.warn("timeout!")
 
     def init_packet_handlers(self):
@@ -72,17 +74,11 @@ class GObjectXpraClient(XpraClientBase, gobject.GObject):
     def gobject_init(self):
         gobject.threads_init()
 
-    def connect_with_timeout(self, conn):
-        self.setup_connection(conn)
-        if conn.timeout>0:
-            glib.timeout_add((conn.timeout + EXTRA_TIMEOUT) * 1000, self.timeout)
-        glib.idle_add(self.send_hello)
-
     def run(self):
         XpraClientBase.run(self)
         self.glib_mainloop = glib.MainLoop()
         self.glib_mainloop.run()
-        return  self.exit_code
+        return self.exit_code
 
     def make_hello(self):
         capabilities = XpraClientBase.make_hello(self)
@@ -103,6 +99,7 @@ class CommandConnectClient(GObjectXpraClient):
         Utility superclass for clients that only send one command
         via the hello packet.
     """
+    COMMAND_TIMEOUT = EXTRA_TIMEOUT
 
     def __init__(self, conn, opts):
         GObjectXpraClient.__init__(self)
@@ -115,19 +112,25 @@ class CommandConnectClient(GObjectXpraClient):
         #so don't try probing for printers, etc
         self.file_transfer = False
         self.printing = False
-
-    def make_hello(self):
-        capabilities = GObjectXpraClient.make_hello(self)
         #don't bother with many of these things for one-off commands:
         for x in ("ui_client", "wants_aliases", "wants_encodings",
                   "wants_versions", "wants_features", "wants_sound", "windows"):
-            capabilities[x] = False
-        return capabilities
+            self.hello_extra[x] = False
 
-    def _process_connection_lost(self, packet):
+    def connect_with_timeout(self, conn):
+        self.setup_connection(conn)
+        if conn.timeout>0:
+            glib.timeout_add((conn.timeout + self.COMMAND_TIMEOUT) * 1000, self.timeout)
+        glib.idle_add(self.send_hello)
+
+    def _process_connection_lost(self, _packet):
         #override so we don't log a warning
         #"command clients" are meant to exit quickly by losing the connection
-        self.quit(EXIT_OK)
+        p = self._protocol
+        if p and p.input_packetcount==0:
+            self.quit(EXIT_CONNECTION_LOST)
+        else:
+            self.quit(EXIT_OK)
 
     def server_connection_established(self):
         #don't bother parsing the network caps:
@@ -136,6 +139,7 @@ class CommandConnectClient(GObjectXpraClient):
         log("server_capabilities: %s", self.server_capabilities)
         log("protocol state: %s", self._protocol.save_state())
         self.do_command()
+        return True
 
     def do_command(self):
         raise NotImplementedError()
@@ -151,7 +155,28 @@ class SendCommandConnectClient(CommandConnectClient):
     def server_connection_established(self):
         assert self.parse_encryption_capabilities(), "encryption failure"
         assert self.parse_network_capabilities(), "network capabilities failure"
-        CommandConnectClient.server_connection_established(self)
+        return CommandConnectClient.server_connection_established(self)
+
+
+class HelloRequestClient(SendCommandConnectClient):
+    """
+        Utility superclass for clients that send a server request
+        as part of the hello packet.
+    """
+
+    def make_hello_base(self):
+        caps = CommandConnectClient.make_hello_base(self)
+        caps.update(self.hello_request())
+        return caps
+
+    def timeout(self, *_args):
+        self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
+
+    def hello_request(self):
+        raise NotImplementedError()
+
+    def do_command(self):
+        self.quit(EXIT_OK)
 
 
 class ScreenshotXpraClient(CommandConnectClient):
@@ -163,13 +188,15 @@ class ScreenshotXpraClient(CommandConnectClient):
     def __init__(self, conn, opts, screenshot_filename):
         self.screenshot_filename = screenshot_filename
         CommandConnectClient.__init__(self, conn, opts)
+        self.hello_extra["screenshot_request"] = True
+        self.hello_extra["request"] = "screenshot"
 
-    def timeout(self, *args):
+    def timeout(self, *_args):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the screenshot")
 
     def _process_screenshot(self, packet):
         (w, h, encoding, _, img_data) = packet[1:6]
-        assert encoding=="png"
+        assert encoding==b"png", "expected png screenshot data but got %s" % bytestostr(encoding)
         if len(img_data)==0:
             self.warn_and_quit(EXIT_OK, "screenshot is empty and has not been saved (maybe there are no windows or they are not currently shown)")
             return
@@ -181,58 +208,86 @@ class ScreenshotXpraClient(CommandConnectClient):
         GObjectXpraClient.init_packet_handlers(self)
         self._ui_packet_handlers["screenshot"] = self._process_screenshot
 
-    def make_hello(self):
-        capabilities = GObjectXpraClient.make_hello(self)
-        capabilities["screenshot_request"] = True
-        return capabilities
-
 
 class InfoXpraClient(CommandConnectClient):
     """ This client does one thing only:
         it queries the server with an 'info' request
     """
 
-    def timeout(self, *args):
+    def __init__(self, *args):
+        CommandConnectClient.__init__(self, *args)
+        self.hello_extra["info_request"] = True
+        self.hello_extra["request"] = "info"
+        if FLATTEN_INFO>=1:
+            self.hello_extra["info-namespace"] = True
+
+    def timeout(self, *_args):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the info")
 
     def do_command(self):
         if self.server_capabilities:
             if FLATTEN_INFO<2:
                 #compatibility mode:
-                for k in sorted_nicely(self.server_capabilities.keys()):
-                    v = self.server_capabilities.get(k)
-                    if sys.version_info[0]>=3:
+                c = flatten_dict(self.server_capabilities)
+                for k in sorted_nicely(c.keys()):
+                    v = c.get(k)
+                    if PYTHON3:
                         #FIXME: this is a nasty and horrible python3 workaround (yet again)
                         #we want to print bytes as strings without the ugly 'b' prefix..
                         #it assumes that all the strings are raw or in (possibly nested) lists or tuples only
+                        #we assume that all strings we get are utf-8,
+                        #and fallback to the bytestostr hack if that fails
                         def fixvalue(w):
                             if type(w)==bytes:
-                                return bytestostr(w)
+                                try:
+                                    return w.decode("utf-8")
+                                except:
+                                    return bytestostr(w)
                             elif type(w) in (tuple,list):
                                 return type(w)([fixvalue(x) for x in w])
                             return w
                         v = fixvalue(v)
-                    log.info("%s=%s", bytestostr(k), nonl(v))
+                        k = fixvalue(k)
+                    log.info("%s=%s", k, nonl(v))
             else:
                 print_nested_dict(self.server_capabilities)
         self.quit(EXIT_OK)
 
-    def make_hello(self):
-        capabilities = GObjectXpraClient.make_hello(self)
-        log("make_hello() adding info_request to %s", capabilities)
-        capabilities["info_request"] = True
-        if FLATTEN_INFO!=1:
-            capabilities["info-namespace"] = True
-        return capabilities
+class IDXpraClient(InfoXpraClient):
+
+    def __init__(self, *args):
+        InfoXpraClient.__init__(self, *args)
+        self.hello_extra["request"] = "id"
+
 
 class ConnectTestXpraClient(CommandConnectClient):
     """ This client does one thing only:
         it queries the server with an 'info' request
     """
 
-    def timeout(self, *args):
+    def __init__(self, *args, **kwargs):
+        CommandConnectClient.__init__(self, *args)
+        self.value = get_hex_uuid()
+        self.hello_extra.update({
+            "connect_test_request"      : self.value,
+            "request"                   : "connect_test",
+            #tells proxy servers we don't want to connect to the real / new instance:
+            "connect"                   : False,
+            #older servers don't know about connect-test,
+            #pretend that we're interested in info:
+            "info_request"              : True,
+            "info-namespace"            : True,
+            })
+        self.hello_extra.update(kwargs)
+
+    def timeout(self, *_args):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: no server response")
 
+    def _process_connection_lost(self, _packet):
+        #we should always receive a hello back and call do_command,
+        #which sets the correct exit code, landing here is an error:
+        self.quit(EXIT_FAILURE)
+
     def do_command(self):
         if self.server_capabilities:
             ctr = self.server_capabilities.get("connect_test_response")
@@ -244,18 +299,6 @@ class ConnectTestXpraClient(CommandConnectClient):
         else:
             self.quit(EXIT_FAILURE)
 
-    def make_hello(self):
-        self.value = get_hex_uuid()
-        capabilities = GObjectXpraClient.make_hello(self)
-        capabilities.update({
-                             "connect_test_request"     : self.value,
-                             #older servers don't know about connect-test,
-                             #pretend that we're interested in info:
-                             "info_request"             : True,
-                             "info-namespace"           : True,
-                             })
-        return capabilities
-
 
 class MonitorXpraClient(SendCommandConnectClient):
     """ This client does one thing only:
@@ -263,6 +306,12 @@ class MonitorXpraClient(SendCommandConnectClient):
         If the server does not support this feature it exits with an error.
     """
 
+    def __init__(self, *args):
+        SendCommandConnectClient.__init__(self, *args)
+        for x in ("wants_features", "wants_events", "event_request"):
+            self.hello_extra[x] = True
+        self.hello_extra["request"] = "event"
+
     def timeout(self, *args):
         pass
         #self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the info")
@@ -278,35 +327,34 @@ class MonitorXpraClient(SendCommandConnectClient):
         self._packet_handlers["server-event"] = self._process_server_event
         self._packet_handlers["ping"] = self._process_ping
 
-    def make_hello(self):
-        capabilities = SendCommandConnectClient.make_hello(self)
-        log("make_hello() adding info_request to %s", capabilities)
-        capabilities["wants_features"]  = True      #so we can verify that the server supports monitor mode
-        capabilities["wants_events"]    = True      #tell the server we do support server events
-        capabilities["event_request"]   = True      #ask the server to enter this request mode (we're not a proper client)
-        return capabilities
-
     def _process_ping(self, packet):
         echotime = packet[1]
         self.send("ping_echo", echotime, 0, 0, 0, -1)
 
 
-class VersionXpraClient(CommandConnectClient):
+class VersionXpraClient(HelloRequestClient):
     """ This client does one thing only:
         it queries the server for version information and prints it out
     """
 
-    def timeout(self, *args):
-        self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the version")
+    def hello_request(self):
+        return {
+            "version_request"       : True,
+            "request"               : "version",
+            "full-version-request"  : True,
+            }
+
+    def parse_network_capabilities(self, *_args):
+        #don't bother checking anything - this could generate warnings
+        return True
 
     def do_command(self):
-        self.warn_and_quit(EXIT_OK, str(self.server_capabilities.get("version")))
-
-    def make_hello(self):
-        capabilities = GObjectXpraClient.make_hello(self)
-        log("make_hello() adding version_request to %s", capabilities)
-        capabilities["version_request"] = True
-        return capabilities
+        v = self.server_capabilities.get(b"version")
+        if not v:
+            self.warn_and_quit(EXIT_FAILURE, "server did not provide the version information")
+        else:
+            sys.stdout.write("%s\n" % (bytestostr(v),))
+            self.quit(EXIT_OK)
 
 
 class ControlXpraClient(CommandConnectClient):
@@ -315,7 +363,7 @@ class ControlXpraClient(CommandConnectClient):
     def set_command_args(self, command):
         self.command = command
 
-    def timeout(self, *args):
+    def timeout(self, *_args):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not respond")
 
     def do_command(self):
@@ -335,6 +383,7 @@ class ControlXpraClient(CommandConnectClient):
         capabilities = GObjectXpraClient.make_hello(self)
         log("make_hello() adding command request '%s' to %s", self.command, capabilities)
         capabilities["command_request"] = self.command
+        capabilities["request"] = "command"
         return capabilities
 
 
@@ -375,7 +424,7 @@ class PrintClient(SendCommandConnectClient):
     def client_type(self):
         return "Python/GObject/Print"
 
-    def timeout(self, *args):
+    def timeout(self, *_args):
         self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not respond")
 
     def do_command(self):
@@ -394,61 +443,108 @@ class PrintClient(SendCommandConnectClient):
         capabilities = SendCommandConnectClient.make_hello(self)
         capabilities["wants_features"] = True   #so we know if printing is supported or not
         capabilities["print_request"] = True    #marker to skip full setup
+        capabilities["request"] = "print"
         return capabilities
 
 
-class ExitXpraClient(SendCommandConnectClient):
+class ExitXpraClient(HelloRequestClient):
     """ This client does one thing only:
         it asks the server to terminate (like stop),
         but without killing the Xvfb or clients.
     """
 
-    def timeout(self, *args):
-        self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
-
-    def make_hello(self):
-        capabilities = SendCommandConnectClient.make_hello(self)
-        capabilities["exit_request"] = True
-        return capabilities
+    def hello_request(self):
+        return {
+            "exit_request"  : True,
+            "request"       : "exit",
+            }
 
     def do_command(self):
         self.idle_add(self.send, "exit-server")
 
 
-class StopXpraClient(SendCommandConnectClient):
+class StopXpraClient(HelloRequestClient):
     """ stop a server """
 
-    def make_hello(self):
-        #used for telling the proxy server we want "stop"
-        #(as waiting for the hello back would be too late)
-        capabilities = SendCommandConnectClient.make_hello(self)
-        capabilities["stop_request"] = True
-        return capabilities
-
-    def timeout(self, *args):
-        self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
+    def hello_request(self):
+        return {
+            "stop_request"  : True,
+            "request"       : "stop",
+            }
 
     def do_command(self):
-        self.idle_add(self.send, "shutdown-server")
+        if not self.server_client_shutdown:
+            log.error("Error: cannot shutdown this server")
+            log.error(" the feature is disable on the server")
+            self.quit(EXIT_FAILURE)
+            return
+        self.idle_add(self.send_shutdown_server)
         #not exiting the client here,
         #the server should send us the shutdown disconnection message anyway
         #and if not, we will then hit the timeout to tell us something went wrong
 
-class DetachXpraClient(SendCommandConnectClient):
+
+class DetachXpraClient(HelloRequestClient):
     """ run the detach subcommand """
 
-    def make_hello(self):
-        #used for telling the proxy server we want "detach"
-        #(older versions ignore this flag and detach because this is a new valid connection
-        # but this breaks if sharing is enabled!)
-        capabilities = SendCommandConnectClient.make_hello(self)
-        capabilities["detach_request"] = True
-        return capabilities
-
-    def timeout(self, *args):
-        self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
+    def hello_request(self):
+        return {
+            "detach_request"    : True,
+            "request"           : "detach",
+            }
 
     def do_command(self):
         self.idle_add(self.send, "disconnect", DONE, "detaching")
         #not exiting the client here,
         #the server should disconnect us with the response
+
+class WaitForDisconnectXpraClient(DetachXpraClient):
+    """ we just want the connection to close """
+
+    def _process_disconnect(self, _packet):
+        self.quit(EXIT_OK)
+
+
+class RequestStartClient(HelloRequestClient):
+    """ request the system proxy server to start a new session for us """
+    #wait longer for this command to return:
+    from xpra.scripts.main import WAIT_SERVER_TIMEOUT
+    COMMAND_TIMEOUT = EXTRA_TIMEOUT+WAIT_SERVER_TIMEOUT
+
+    def hello_request(self):
+        return {
+            "start-new-session" : self.start_new_session,
+            #tells proxy servers we don't want to connect to the real / new instance:
+            "connect"                   : False,
+            }
+
+    def server_connection_established(self):
+        #the server should respond with the display chosen
+        log("server_connection_established() exit_code=%s", self.exit_code)
+        display = self.server_capabilities.get("display")
+        if display:
+            try:
+                mode = self.server_capabilities.get("mode")
+                session_type = {
+                    "start"         : "seamless ",
+                    "start-desktop" : "desktop ",
+                    "shadow"        : "shadow ",
+                    }.get(mode, "")
+                sys.stderr.write("%ssession now available on display %s\n" % (session_type, display))
+                sys.stderr.flush()
+                if POSIX and not OSX and self.displayfd>0 and display and display.startswith(b":"):
+                    from xpra.platform.displayfd import write_displayfd
+                    log("writing display %s to displayfd=%s", display, self.displayfd)
+                    write_displayfd(self.displayfd, display[1:])
+            except:
+                log("server_connection_established()", exc_info=True)
+        if not self.exit_code:
+            self.quit(0)
+        return True
+
+    def __init__(self, conn, opts):
+        HelloRequestClient.__init__(self, conn, opts)
+        try:
+            self.displayfd = int(opts.displayfd)
+        except:
+            self.displayfd = 0
