@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2010-2016 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2017 Antoine Martin <antoine@devloop.org.uk>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -11,8 +11,9 @@ import os
 #before we import xpra.platform
 import platform as python_platform
 assert python_platform
-from xpra.util import AdHocStruct
-from xpra.os_util import WIN32, OSX
+from xpra.util import csv
+from xpra.os_util import (WIN32, OSX, PYTHON2, osexpand, getuid, getgid, get_username_for_uid,
+    is_CentOS, is_RedHat, is_Fedora, is_Debian, is_Ubuntu, getUbuntuVersion)
 
 def warn(msg):
     sys.stderr.write(msg+"\n")
@@ -40,9 +41,6 @@ DEFAULT_PULSEAUDIO = None   #auto
 if OSX or WIN32:
     DEFAULT_PULSEAUDIO = False
 
-POSIX = os.name=="posix"
-PYTHON3 = sys.version_info[0]>=3
-
 
 _has_sound_support = None
 def has_sound_support():
@@ -58,6 +56,9 @@ def has_sound_support():
 
 def get_xorg_bin():
     # Detect Xorg Binary
+    if os.uname()[4].startswith("arm") and is_Debian() and os.path.exists("/usr/bin/Xorg"):
+        #Raspbian breaks if we use a different binary..
+        return "/usr/bin/Xorg"
     for p in (
               "/usr/libexec/Xorg",              #fedora 22+
               "/usr/lib/xorg/Xorg",             #ubuntu 16.10
@@ -74,10 +75,19 @@ def get_xorg_bin():
     return None
 
 
+def get_Xdummy_confdir():
+    from xpra.platform.xposix.paths import get_runtime_dir
+    xrd = get_runtime_dir()
+    if xrd:
+        base = "${XDG_RUNTIME_DIR}/xpra"
+    else:
+        base = "${HOME}/.xpra"
+    return base+"/xorg.conf.d/$PID"
+
 def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/etc/xpra/xorg.conf"):
     cmd = [xorg_cmd]    #ie: ["Xorg"] or ["xpra_Xdummy"] or ["./install/bin/xpra_Xdummy"]
     cmd += [
-          "-noreset",
+          "-noreset", "-novtswitch",
           "-nolisten", "tcp",
           "+extension", "GLX",
           "+extension", "RANDR",
@@ -86,13 +96,14 @@ def get_Xdummy_command(xorg_cmd="Xorg", log_dir="${XPRA_LOG_DIR}", xorg_conf="/e
           "-logfile", "%s/Xorg.${DISPLAY}.log" % log_dir,
           #must be specified with some Xorg versions (ie: arch linux)
           #this directory can store xorg config files, it does not need to be created:
-          "-configdir", "${HOME}/.xpra/xorg.conf.d",
+          "-configdir", get_Xdummy_confdir(),
           "-config", xorg_conf
           ]
     return cmd
 
 def get_Xvfb_command():
     cmd = ["Xvfb",
+           "+extension", "GLX",
            "+extension", "Composite",
            "-screen", "0", "5760x2560x24+32",
            #better than leaving to vfb after a resize?
@@ -151,7 +162,6 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None
     else:
         debug("Xdummy support unspecified, will try to detect")
 
-    from xpra.os_util import is_Ubuntu, getUbuntuVersion
     if is_Ubuntu():
         rnum = getUbuntuVersion()
         if rnum==[16, 10]:
@@ -162,12 +172,6 @@ def detect_xvfb_command(conf_dir="/etc/xpra/", bin_dir=None, Xdummy_ENABLED=None
 
 
 def OpenGL_safety_check():
-    #Ubuntu 12.04 will just crash on you if you try:
-    from xpra.os_util import is_Ubuntu, getUbuntuVersion
-    if is_Ubuntu():
-        rnum = getUbuntuVersion()
-        if rnum<=[12, 4]:
-            return "Ubuntu %s is too buggy" % rnum
     #try to detect VirtualBox:
     #based on the code found here:
     #http://spth.virii.lu/eof2/articles/WarGame/vboxdetect.html
@@ -192,11 +196,11 @@ def OpenGL_safety_check():
             return "VirtualBox is present (VBoxMiniRdrDN)"
     return None
 
-OPENGL_DEFAULT = None       #will auto-detect by probing
+OPENGL_DEFAULT = "auto"       #will auto-detect by probing
 def get_opengl_default():
     global OPENGL_DEFAULT
     if OpenGL_safety_check() is not None:
-        OPENGL_DEFAULT = False
+        OPENGL_DEFAULT = "no"
     return OPENGL_DEFAULT
 
 
@@ -259,7 +263,11 @@ def read_config(conf_file):
     if not os.path.isfile(conf_file):
         debug("read_config(%s) is not a file!", conf_file)
         return d
-    with open(conf_file, "rU") as f:
+    if PYTHON2:
+        mode = "rU"
+    else:
+        mode = "r"
+    with open(conf_file, mode) as f:
         lines = []
         no = 0
         for line in f:
@@ -323,7 +331,7 @@ def conf_files(conf_dir, xpra_conf_filename=DEFAULT_XPRA_CONF_FILENAME):
     #look for conf.d subdirectory:
     conf_d_dir = os.path.join(cdir, "conf.d")
     if os.path.exists(conf_d_dir) and os.path.isdir(conf_d_dir):
-        for f in os.listdir(conf_d_dir):
+        for f in sorted(os.listdir(conf_d_dir)):
             if f.endswith(".conf"):
                 conf_file = os.path.join(conf_d_dir, f)
                 if os.path.isfile(conf_file):
@@ -349,7 +357,7 @@ def read_xpra_conf(conf_dir, xpra_conf_filename=DEFAULT_XPRA_CONF_FILENAME):
         d.update(cd)
     return d
 
-def read_xpra_defaults():
+def read_xpra_defaults(username=None, uid=None, gid=None):
     """
         Reads the global <xpra_conf_filename> from the <conf_dir>
         and then the user-specific one.
@@ -369,12 +377,12 @@ def read_xpra_defaults():
     # * user config            (ie: "~/.xpra/" on all Posix, including OSX)
     #                          (ie: "C:\Documents and Settings\Username\Application Data\Xpra" with XP)
     #                          (ie: "C:\Users\<user name>\AppData\Roaming" with Visa onwards)
-    dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs()
+    dirs = get_default_conf_dirs() + get_system_conf_dirs() + get_user_conf_dirs(uid)
     defaults = {}
     for d in dirs:
         if not d:
             continue
-        ad = os.path.expanduser(d)
+        ad = osexpand(d, actual_username=username, uid=uid, gid=gid)
         if not os.path.exists(ad):
             debug("read_xpra_defaults: skipping %s", ad)
             continue
@@ -401,7 +409,7 @@ def may_create_user_config(xpra_conf_filename=DEFAULT_XPRA_CONF_FILENAME):
                 try:
                     if not os.path.exists(ad):
                         os.makedirs(ad, int('700', 8))
-                    with open(conf_file, 'wb') as f:
+                    with open(conf_file, 'w') as f:
                         f.write("# xpra user configuration file\n")
                         f.write("# place your custom settings in this file\n")
                         f.write("# they will take precedence over the system default ones.\n")
@@ -426,19 +434,15 @@ OPTIONS_VALIDATION = {}
 OPTION_TYPES = {
                     #string options:
                     "encoding"          : str,
+                    "opengl"            : str,
                     "title"             : str,
                     "username"          : str,
                     "password"          : str,
-                    "auth"              : str,
-                    "vsock-auth"        : str,
-                    "tcp-auth"          : str,
-                    "ssl-auth"          : str,
                     "wm-name"           : str,
                     "session-name"      : str,
                     "dock-icon"         : str,
                     "tray-icon"         : str,
                     "window-icon"       : str,
-                    "password-file"     : str,
                     "keyboard-raw"      : bool,
                     "keyboard-layout"   : str,
                     "keyboard-layouts"  : list,
@@ -451,6 +455,7 @@ OPTION_TYPES = {
                     "remote-clipboard"  : str,
                     "local-clipboard"   : str,
                     "pulseaudio-command": str,
+                    "bandwidth-limit"   : str,
                     "tcp-encryption"    : str,
                     "tcp-encryption-keyfile": str,
                     "encryption"        : str,
@@ -460,6 +465,8 @@ OPTION_TYPES = {
                     "ssh"               : str,
                     "systemd-run"       : str,
                     "systemd-run-args"  : str,
+                    "system-proxy-socket" : str,
+                    "chdir"             : str,
                     "xvfb"              : str,
                     "socket-dir"        : str,
                     "mmap"              : str,
@@ -489,6 +496,14 @@ OPTION_TYPES = {
                     "exec-wrapper"      : str,
                     "dbus-launch"       : str,
                     "webcam"            : str,
+                    "mousewheel"        : str,
+                    "input-devices"     : str,
+                    "shortcut-modifiers": str,
+                    "open-files"        : str,
+                    "open-url"          : str,
+                    "file-transfer"     : str,
+                    "printing"          : str,
+                    "challenge-handlers": str,
                     #ssl options:
                     "ssl"               : str,
                     "ssl-key"           : str,
@@ -504,6 +519,7 @@ OPTION_TYPES = {
                     "ssl-server-hostname" : str,
                     "ssl-options"       : str,
                     #int options:
+                    "displayfd"         : int,
                     "pings"             : int,
                     "quality"           : int,
                     "min-quality"       : int,
@@ -517,10 +533,16 @@ OPTION_TYPES = {
                     "server-idle-timeout" : int,
                     "sync-xvfb"         : int,
                     "pixel-depth"       : int,
+                    "uid"               : int,
+                    "gid"               : int,
+                    "min-port"          : int,
+                    "rfb-upgrade"       : int,
                     #float options:
                     "auto-refresh-delay": float,
                     #boolean options:
                     "daemon"            : bool,
+                    "start-via-proxy"   : bool,
+                    "attach"            : bool,
                     "use-display"       : bool,
                     "fake-xinerama"     : bool,
                     "resize_display"    : bool,
@@ -536,23 +558,22 @@ OPTION_TYPES = {
                     "xsettings"         : bool,
                     "system-tray"       : bool,
                     "sharing"           : bool,
+                    "lock"              : bool,
                     "delay-tray"        : bool,
                     "windows"           : bool,
+                    "terminate-children": bool,
                     "exit-with-children": bool,
                     "exit-with-client"  : bool,
                     "exit-ssh"          : bool,
                     "dbus-control"      : bool,
                     "av-sync"           : bool,
-                    "opengl"            : bool,
                     "mdns"              : bool,
-                    "file-transfer"     : bool,
-                    "printing"          : bool,
-                    "open-files"        : bool,
                     "swap-keys"         : bool,
                     "start-new-commands": bool,
                     "proxy-start-sessions": bool,
                     "desktop-fullscreen": bool,
                     "global-menus"      : bool,
+                    "forward-xdg-open"  : bool,
                     #arrays of strings:
                     "pulseaudio-configure-commands" : list,
                     "socket-dirs"       : list,
@@ -576,37 +597,149 @@ OPTION_TYPES = {
                     "bind"              : list,
                     "bind-vsock"        : list,
                     "bind-tcp"          : list,
+                    "bind-udp"          : list,
+                    "bind-ws"           : list,
+                    "bind-wss"          : list,
                     "bind-ssl"          : list,
+                    "bind-rfb"          : list,
+                    "auth"              : list,
+                    "vsock-auth"        : list,
+                    "tcp-auth"          : list,
+                    "udp-auth"          : list,
+                    "ws-auth"           : list,
+                    "wss-auth"          : list,
+                    "ssl-auth"          : list,
+                    "rfb-auth"          : list,
+                    "password-file"     : list,
                     "start-env"         : list,
                     "env"               : list,
                }
 
+#in the options list, available in session files,
+#but not on the command line:
+NON_COMMAND_LINE_OPTIONS = [
+    "mode",
+    "proxy-video-encoders",
+    "display",
+    ]
+
+START_COMMAND_OPTIONS = [
+    "start", "start-child",
+    "start-after-connect", "start-child-after-connect",
+    "start-on-connect", "start-child-on-connect",
+    ]
+BIND_OPTIONS = ["bind", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-vsock", "bind-rfb"]
+
+#keep track of the options added since v1,
+#so we can generate command lines that work with older supported versions:
+OPTIONS_ADDED_SINCE_V1 = ["attach", "open-files", "open-url", "pixel-depth", "uid", "gid", "chdir", "min-port", "rfb-upgrade", "bandwidth-limit",
+                    "forward-xdg-open",
+                          ]
+OPTIONS_COMPAT_NAMES = {
+    "--compression_level=" : "-z"
+    }
+
+CLIENT_OPTIONS = ["title", "username", "password", "session-name",
+                  "dock-icon", "tray-icon", "window-icon",
+                  "clipboard", "clipboard-direction", "clipboard-filter-file",
+                  "remote-clipboard", "local-clipboard",
+                  "tcp-encryption", "tcp-encryption-keyfile", "encryption",  "encryption-keyfile",
+                  "systemd-run", "systemd-run-args",
+                  "socket-dir", "socket-dirs",
+                  "border", "window-close", "max-size", "desktop-scaling",
+                  "file-transfer", "file-size-limit", "download-path", "open-command", "open-files", "printing", "open-url",
+                  "challenge-handlers",
+                  "dbus-proxy",
+                  "remote-logging",
+                  "lpadmin", "lpinfo",
+                  "debug",
+                  "microphone", "speaker", "sound-source",
+                  "microphone-codec", "speaker-codec",
+                  "mmap", "encodings", "encoding",
+                  "quality", "min-quality", "speed", "min-speed",
+                  "compression_level",
+                  "dpi", "video-scaling", "auto-refresh-delay",
+                  "webcam", "mousewheel", "input-devices", "shortcut-modifiers", "pings",
+                  "tray", "keyboard-sync", "cursors", "bell", "notifications",
+                  "xsettings", "system-tray", "sharing", "lock",
+                  "delay-tray", "windows", "readonly",
+                  "av-sync", "swap-keys",
+                  "opengl",
+                  "start-new-commands",
+                  "desktop-fullscreen", "global-menus",
+                  "video-encoders", "csc-modules", "video-decoders",
+                  "compressors", "packet-encoders",
+                  "key-shortcut",
+                  "env"]
+
+CLIENT_ONLY_OPTIONS = ["username", "swap-keys", "dock-icon",
+                       "tray", "delay-tray", "tray-icon"]
+
+#options that clients can pass to the proxy
+#and which will be forwarded to the new proxy instance process:
+PROXY_START_OVERRIDABLE_OPTIONS = [
+    "env", "start-env", "chdir",
+    "dpi",
+    "encoding", "encodings",
+    "quality", "min-quality", "speed", "min-speed",
+    #"auto-refresh-delay",    float!
+    #no corresponding command line option:
+    #"wm-name", "download-path",
+    "compression_level", "video-scaling",
+    "title", "session-name",
+    "clipboard", "clipboard-direction", "clipboard-filter-file",
+    "input-method",
+    "microphone", "speaker", "sound-source", "pulseaudio",
+    "idle-timeout", "server-idle-timeout",
+    "use-display", "fake-xinerama", "resize_display", "dpi", "pixel-depth",
+    "readonly", "keyboard-sync", "cursors", "bell", "notifications", "xsettings",
+    "system-tray", "sharing", "lock", "windows", "webcam", "html",
+    "terminate-children", "exit-with-children", "exit-with-client",
+    "av-sync", "global-menus",
+    "forward-xdg-open",
+    "printing", "file-transfer", "open-command", "open-files", "open-url", "start-new-commands",
+    "mmap", "mmap-group", "mdns",
+    "auth", "vsock-auth", "tcp-auth", "udp-auth", "ws-auth", "wss-auth", "ssl-auth", "rfb-auth",
+    "bind", "bind-vsock", "bind-tcp", "bind-udp", "bind-ssl", "bind-ws", "bind-wss", "bind-rfb",
+    "rfb-upgrade", "bandwidth-limit",
+    "start", "start-child",
+    "start-after-connect", "start-child-after-connect",
+    "start-on-connect", "start-child-on-connect",
+    ]
+tmp = os.environ.get("XPRA_PROXY_START_OVERRIDABLE_OPTIONS", "")
+if tmp:
+    PROXY_START_OVERRIDABLE_OPTIONS = tmp.split(",")
+del tmp
+
+
 def get_default_key_shortcuts():
     return [shortcut for e,shortcut in [
                (True,   "Control+Menu:toggle_keyboard_grab"),
                (True,   "Shift+Menu:toggle_pointer_grab"),
-               (True,   "Meta+Shift+F1:show_menu"),
-               (True,   "Meta+Shift+F2:show_start_new_command"),
-               (True,   "Meta+Shift+F3:show_bug_report"),
-               (True,   "Meta+Shift+F4:quit"),
-               (True,   "Meta+Shift+F5:increase_quality"),
-               (True,   "Meta+Shift+F6:decrease_quality"),
-               (True,   "Meta+Shift+F7:increase_speed"),
-               (True,   "Meta+Shift+F8:decrease_speed"),
-               (True,   "Meta+Shift+F10:magic_key"),
-               (True,   "Meta+Shift+F11:show_session_info"),
-               (True,   "Meta+Shift+F12:toggle_debug"),
-               (True,   "Meta+Shift+plus:scaleup"),
-               (OSX,    "Meta+Shift+plusminus:scaleup"),
-               (True,   "Meta+Shift+minus:scaledown"),
-               (True,   "Meta+Shift+underscore:scaledown"),
-               (OSX,    "Meta+Shift+emdash:scaledown"),
-               (True,   "Meta+Shift+KP_Add:scaleup"),
-               (True,   "Meta+Shift+KP_Subtract:scaledown"),
-               (True,   "Meta+Shift+KP_Multiply:scalereset"),
-               (True,   "Meta+Shift+bar:scalereset"),
-               (True,   "Meta+Shift+question:scalingoff"),
-               (OSX,    "Meta+Shift+degree:scalereset")]
+               (not OSX,"Shift+F11:toggle_fullscreen"),
+               (OSX,    "Control+F11:toggle_fullscreen"),
+               (True,   "#+F1:show_menu"),
+               (True,   "#+F2:show_start_new_command"),
+               (True,   "#+F3:show_bug_report"),
+               (True,   "#+F4:quit"),
+               (True,   "#+F5:increase_quality"),
+               (True,   "#+F6:decrease_quality"),
+               (True,   "#+F7:increase_speed"),
+               (True,   "#+F8:decrease_speed"),
+               (True,   "#+F10:magic_key"),
+               (True,   "#+F11:show_session_info"),
+               (True,   "#+F12:toggle_debug"),
+               (True,   "#+plus:scaleup"),
+               (OSX,    "#+plusminus:scaleup"),
+               (True,   "#+minus:scaledown"),
+               (True,   "#+underscore:scaledown"),
+               (OSX,    "#+emdash:scaledown"),
+               (True,   "#+KP_Add:scaleup"),
+               (True,   "#+KP_Subtract:scaledown"),
+               (True,   "#+KP_Multiply:scalereset"),
+               (True,   "#+bar:scalereset"),
+               (True,   "#+question:scalingoff"),
+               (OSX,    "#+degree:scalereset")]
                  if e]
 
 def get_default_systemd_run():
@@ -614,12 +747,30 @@ def get_default_systemd_run():
         return "no"
     #don't use systemd-run on CentOS / RedHat
     #(it causes failures with "Failed to create bus connection: No such file or directory")
-    from xpra.os_util import load_binary_file, strtobytes
-    data = strtobytes(load_binary_file("/etc/redhat-release") or "")
-    if data and (data.find(b"RedHat")>=0 or data.find(b"CentOS")>=0):
+    if is_CentOS() or is_RedHat():
         return "no"
+    if is_Fedora():
+        return "no"
+    #systemd-run was previously broken in Fedora 26:
+    #https://github.com/systemd/systemd/issues/3388
+    #but with newer kernels, it is working again..
     return "auto"
 
+def get_default_pulseaudio_command():
+    if WIN32 or OSX:
+        return []
+    cmd = ["pulseaudio", "--start", "-n", "--daemonize=false", "--system=false",
+                                    "--exit-idle-time=-1", "--load=module-suspend-on-idle",
+                                    "'--load=module-null-sink sink_name=\"Xpra-Speaker\" sink_properties=device.description=\"Xpra\\ Speaker\"'",
+                                    "'--load=module-null-sink sink_name=\"Xpra-Microphone\" sink_properties=device.description=\"Xpra\\ Microphone\"'",
+                                    "'--load=module-native-protocol-unix socket=$XPRA_PULSE_SERVER'",
+                                    "--load=module-dbus-protocol",
+                                    "--load=module-x11-publish",
+                                    "--log-level=2", "--log-target=stderr"]
+    if not is_Ubuntu() or getUbuntuVersion()>[16, 4]:
+        cmd.append("--enable-memfd=no")
+    return cmd
+
 
 GLOBAL_DEFAULTS = None
 #lowest common denominator here
@@ -628,8 +779,8 @@ def get_defaults():
     global GLOBAL_DEFAULTS
     if GLOBAL_DEFAULTS is not None:
         return GLOBAL_DEFAULTS
-    from xpra.platform.features import DEFAULT_SSH_COMMAND, OPEN_COMMAND, DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS, DEFAULT_PULSEAUDIO_COMMAND, \
-                                        DEFAULT_ENV, CAN_DAEMONIZE
+    from xpra.platform.features import DEFAULT_SSH_COMMAND, OPEN_COMMAND, DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS, \
+                                        DEFAULT_ENV, CAN_DAEMONIZE, SYSTEM_PROXY_SOCKET
     from xpra.platform.paths import get_download_dir, get_remote_run_xpra_scripts
     try:
         from xpra.platform.info import get_username
@@ -672,20 +823,15 @@ def get_defaults():
         ssl_protocol = "SSLv23"
 
     GLOBAL_DEFAULTS = {
-                    "encoding"          : "",
+                    "encoding"          : "auto",
                     "title"             : "@title@ on @client-machine@",
                     "username"          : username,
                     "password"          : "",
-                    "auth"              : "",
-                    "vsock-auth"        : "",
-                    "tcp-auth"          : "",
-                    "ssl-auth"          : "",
                     "wm-name"           : DEFAULT_NET_WM_NAME,
                     "session-name"      : "",
                     "dock-icon"         : "",
                     "tray-icon"         : "",
                     "window-icon"       : "",
-                    "password-file"     : "",
                     "keyboard-raw"      : False,
                     "keyboard-layout"   : "",
                     "keyboard-layouts"  : [],
@@ -697,7 +843,8 @@ def get_defaults():
                     "clipboard-filter-file" : "",
                     "remote-clipboard"  : "CLIPBOARD",
                     "local-clipboard"   : "CLIPBOARD",
-                    "pulseaudio-command": " ".join(DEFAULT_PULSEAUDIO_COMMAND),
+                    "pulseaudio-command": " ".join(get_default_pulseaudio_command()),
+                    "bandwidth-limit"   : "auto",
                     "encryption"        : "",
                     "tcp-encryption"    : "",
                     "encryption-keyfile": "",
@@ -706,7 +853,9 @@ def get_defaults():
                     "ssh"               : DEFAULT_SSH_COMMAND,
                     "systemd-run"       : get_default_systemd_run(),
                     "systemd-run-args"  : "",
+                    "system-proxy-socket" : SYSTEM_PROXY_SOCKET,
                     "xvfb"              : " ".join(xvfb),
+                    "chdir"             : "",
                     "socket-dir"        : "",
                     "log-dir"           : "auto",
                     "log-file"          : "$DISPLAY.log",
@@ -717,7 +866,7 @@ def get_defaults():
                     "display"           : "",
                     "tcp-proxy"         : "",
                     "download-path"     : get_download_dir(),
-                    "open-command"      : OPEN_COMMAND,
+                    "open-command"      : " ".join(OPEN_COMMAND),
                     "remote-logging"    : "both",
                     "lpadmin"           : "/usr/sbin/lpadmin",
                     "lpinfo"            : "/usr/sbin/lpinfo",
@@ -732,6 +881,14 @@ def get_defaults():
                     "exec-wrapper"      : "",
                     "dbus-launch"       : "dbus-launch --close-stderr",
                     "webcam"            : ["auto", "no"][OSX],
+                    "mousewheel"        : "on",
+                    "input-devices"     : "auto",
+                    "shortcut-modifiers": "auto",
+                    "open-files"        : "auto",
+                    "open-url"          : "auto",
+                    "file-transfer"     : "auto",
+                    "printing"          : "yes",
+                    "challenge-handlers": "all",
                     #ssl options:
                     "ssl"               : "auto",
                     "ssl-key"           : "",
@@ -757,9 +914,15 @@ def get_defaults():
                     "idle-timeout"      : 0,
                     "server-idle-timeout" : 0,
                     "sync-xvfb"         : 0,
-                    "pixel-depth"       : 24,
+                    "pixel-depth"       : 0,
+                    "uid"               : getuid(),
+                    "gid"               : getgid(),
+                    "min-port"          : 1024,
+                    "rfb-upgrade"       : 5,
                     "auto-refresh-delay": 0.15,
                     "daemon"            : CAN_DAEMONIZE,
+                    "start-via-proxy"   : None,
+                    "attach"            : None,
                     "use-display"       : False,
                     "fake-xinerama"     : not OSX and not WIN32,
                     "resize-display"    : not OSX and not WIN32,
@@ -772,15 +935,18 @@ def get_defaults():
                     "microphone"        : ["disabled", "off"][has_sound_support()],
                     "readonly"          : False,
                     "keyboard-sync"     : True,
+                    "displayfd"         : 0,
                     "pings"             : 5,
                     "cursors"           : True,
                     "bell"              : True,
                     "notifications"     : True,
                     "xsettings"         : not OSX and not WIN32,
                     "system-tray"       : True,
-                    "sharing"           : False,
+                    "sharing"           : None,
+                    "lock"              : None,
                     "delay-tray"        : False,
                     "windows"           : True,
+                    "terminate-children": False,
                     "exit-with-children": False,
                     "exit-with-client"  : False,
                     "start-after-connect": False,
@@ -791,12 +957,10 @@ def get_defaults():
                     "dbus-control"      : not WIN32 and not OSX,
                     "opengl"            : get_opengl_default(),
                     "mdns"              : not WIN32,
-                    "file-transfer"     : True,
-                    "printing"          : True,
-                    "open-files"        : False,
                     "swap-keys"         : OSX,  #only used on osx
                     "desktop-fullscreen": False,
                     "global-menus"      : True,
+                    "forward-xdg-open"  : True,
                     "pulseaudio-configure-commands"  : [" ".join(x) for x in DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS],
                     "socket-dirs"       : [],
                     "remote-xpra"       : get_remote_run_xpra_scripts(),
@@ -813,7 +977,20 @@ def get_defaults():
                     "bind"              : bind_dirs,
                     "bind-vsock"        : [],
                     "bind-tcp"          : [],
+                    "bind-udp"          : [],
+                    "bind-ws"           : [],
+                    "bind-wss"          : [],
                     "bind-ssl"          : [],
+                    "bind-rfb"          : [],
+                    "auth"              : [],
+                    "vsock-auth"        : [],
+                    "tcp-auth"          : [],
+                    "udp-auth"          : [],
+                    "ws-auth"           : [],
+                    "wss-auth"          : [],
+                    "ssl-auth"          : [],
+                    "rfb-auth"          : [],
+                    "password-file"     : [],
                     "start"             : [],
                     "start-child"       : [],
                     "start-after-connect"       : [],
@@ -833,19 +1010,22 @@ NO_FILE_OPTIONS = ["daemon"]
 
 TRUE_OPTIONS = ["yes", "true", "1", "on", True]
 FALSE_OPTIONS = ["no", "false", "0", "off", False]
-def parse_bool(k, v):
+def parse_bool(k, v, auto=None):
     if type(v)==str:
-        v = v.lower()
+        v = v.lower().strip()
     if v in TRUE_OPTIONS:
         return True
     elif v in FALSE_OPTIONS:
         return False
     elif v in ["auto", None]:
         #keep default - which may be None!
-        return None
+        return auto
     else:
-        warn("Warning: cannot parse value '%s' for '%s' as a boolean" % (v, k))
-        return None
+        try:
+            return bool(int(v))
+        except:
+            warn("Warning: cannot parse value '%s' for '%s' as a boolean" % (v, k))
+            return None
 
 def print_bool(k, v, true_str='yes', false_str='no'):
     if type(v)==type(None):
@@ -885,6 +1065,37 @@ def print_number(i, auto_value=0):
         return "auto"
     return str(i)
 
+def parse_with_unit(numtype, v, subunit="bps", min_value=250000):
+    #special case for bandwidth-limit, which can be specified using units:
+    try:
+        v = str(v).lower().strip()
+        import re
+        if not v or v in FALSE_OPTIONS:
+            return 0
+        elif v=="auto":
+            return None
+        else:
+            r = re.match('([0-9\.]*)(.*)', v)
+            assert r
+            f = float(r.group(1))
+            unit = r.group(2).lower()
+            if unit.endswith(subunit):
+                unit = unit[:-len(subunit)]     #ie: 10mbps -> 10m
+            if unit=="b":
+                pass
+            elif unit=="k":
+                f *= 1000
+            elif unit=="m":
+                f *= 1000000
+            elif unit=="g":
+                f *= 1000000000
+            if min_value is not None:
+                assert f>=min_value, "value is too low"
+            return int(f)
+    except Exception as e:
+        raise InitException("invalid value for %s '%s': %s" % (numtype, v, e))
+
+
 def validate_config(d={}, discard=NO_FILE_OPTIONS, extras_types={}, extras_validation={}):
     """
         Validates all the options given in a dict with fields as keys and
@@ -943,9 +1154,11 @@ def validate_config(d={}, discard=NO_FILE_OPTIONS, extras_types={}, extras_valid
     return nd
 
 
-def make_defaults_struct(extras_defaults={}, extras_types={}, extras_validation={}):
+def make_defaults_struct(extras_defaults={}, extras_types={}, extras_validation={}, username="", uid=getuid(), gid=getgid()):
     #populate config with default values:
-    defaults = read_xpra_defaults()
+    if not username and uid:
+        username = get_username_for_uid(uid)
+    defaults = read_xpra_defaults(username, uid, gid)
     return dict_to_validated_config(defaults, extras_defaults, extras_types, extras_validation)
 
 def dict_to_validated_config(d={}, extras_defaults={}, extras_types={}, extras_validation={}):
@@ -957,11 +1170,15 @@ def dict_to_validated_config(d={}, extras_defaults={}, extras_types={}, extras_v
     for k,v in CLONES.items():
         if k in options:
             options[v] = options[k]
-    config = AdHocStruct()
+    config = XpraConfig()
     for k,v in options.items():
         setattr(config, name_to_field(k), v)
     return config
 
+class XpraConfig(object):
+    def __repr__(self):
+        return ("XpraConfig(%s)" % self.__dict__)
+
 
 def fixup_debug_option(value):
     """ backwards compatible parsing of the debug option, which used to be a boolean """
@@ -996,7 +1213,7 @@ def fixup_video_all_or_none(options):
     pvestr  = _csvstr(options.proxy_video_encoders)
     def getlist(strarg, help_txt, all_list):
         if strarg=="help":
-            raise InitInfo("the following %s may be available: %s" % (help_txt, ", ".join(all_list)))
+            raise InitInfo("the following %s may be available: %s" % (help_txt, csv(all_list)))
         elif strarg=="none":
             return []
         elif strarg=="all":
@@ -1063,7 +1280,7 @@ def fixup_compression(options):
         compressors = _nodupes(cstr)
         unknown = [x for x in compressors if x and x not in compression.ALL_COMPRESSORS]
         if unknown:
-            warn("warning: invalid compressor(s) specified: %s" % (", ".join(unknown)))
+            warn("warning: invalid compressor(s) specified: %s" % csv(unknown))
     options.compressors = compressors
 
 def fixup_packetencoding(options):
@@ -1076,7 +1293,7 @@ def fixup_packetencoding(options):
         packet_encoders = _nodupes(pestr)
         unknown = [x for x in packet_encoders if x and x not in packet_encoding.ALL_ENCODERS]
         if unknown:
-            warn("warning: invalid packet encoder(s) specified: %s" % (", ".join(unknown)))
+            warn("warning: invalid packet encoder(s) specified: %s" % csv(unknown))
     options.packet_encoders = packet_encoders
 
 def fixup_keyboard(options):
@@ -1110,9 +1327,6 @@ def fixup_clipboard(options):
         warn(" specify 'to-server', 'to-client' or 'both'")
         options.clipboard_direction = "disabled"
 
-def fixup_bool(options):
-    options.sharing = parse_bool("sharing", options.sharing) or False
-
 def abs_paths(options):
     #convert to absolute paths before we daemonize
     for k in ("clipboard-filter-file",
@@ -1138,7 +1352,6 @@ def fixup_options(options, defaults={}):
     fixup_socketdirs(options, defaults)
     fixup_clipboard(options)
     fixup_keyboard(options)
-    fixup_bool(options)
     abs_paths(options)
     #remote-xpra is meant to be a list, but the user can specify a string using the command line,
     #in which case we replace all the default values with this single entry:
@@ -1154,7 +1367,7 @@ def main():
             if ot==bool and v is None:
                 v = "Auto"
             if type(v)==list:
-                v = ", ".join(str(x) for x in v)
+                v = csv(str(x) for x in v)
             print("* %-32s : %s" % (k, nonl(v)))
     from xpra.platform import program_context
     from xpra.log import enable_color
@@ -1167,8 +1380,6 @@ def main():
                 print(args[0] % args[1:])
             args.remove("-v")
 
-        print("Default Configuration:")
-        print_options(make_defaults_struct())
         if len(args)>0:
             for filename in args:
                 print("")
@@ -1179,6 +1390,9 @@ def main():
                 d = read_config(filename)
                 config = dict_to_validated_config(d)
                 print_options(config)
+        else:
+            print("Default Configuration:")
+            print_options(make_defaults_struct())
 
 
 if __name__ == "__main__":
