@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2013-2016 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2013-2017 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -12,7 +12,7 @@ glib = import_glib()
 glib.threads_init()
 gobject = import_gobject()
 gobject.threads_init()
-from multiprocessing import Queue as MQueue, freeze_support
+from multiprocessing import Queue as MQueue, freeze_support #@UnresolvedImport
 freeze_support()
 
 from xpra.log import Logger
@@ -20,20 +20,21 @@ log = Logger("proxy")
 authlog = Logger("proxy", "auth")
 
 
-from xpra.util import LOGIN_TIMEOUT, AUTHENTICATION_ERROR, PERMISSION_ERROR, SESSION_NOT_FOUND, SERVER_ERROR, repr_ellipsized, print_nested_dict, csv, typedict
-from xpra.os_util import get_username_for_uid, get_groups
+from xpra.util import LOGIN_TIMEOUT, AUTHENTICATION_ERROR, SESSION_NOT_FOUND, SERVER_ERROR, repr_ellipsized, print_nested_dict, csv, envfloat, envbool, typedict
+from xpra.os_util import get_username_for_uid, get_groups, get_home_for_uid, bytestostr, getuid, getgid, WIN32, POSIX
 from xpra.server.proxy.proxy_instance_process import ProxyInstanceProcess
 from xpra.server.server_core import ServerCore
 from xpra.server.control_command import ArgsControlCommand, ControlError
 from xpra.child_reaper import getChildReaper
-from xpra.scripts.config import make_defaults_struct
+from xpra.scripts.config import make_defaults_struct, PROXY_START_OVERRIDABLE_OPTIONS
 from xpra.scripts.main import parse_display_name, connect_to, start_server_subprocess
 from xpra.make_thread import start_thread
 
 
-PROXY_SOCKET_TIMEOUT = float(os.environ.get("XPRA_PROXY_SOCKET_TIMEOUT", "0.1"))
-PROXY_WS_TIMEOUT = float(os.environ.get("XPRA_PROXY_WS_TIMEOUT", "1.0"))
+PROXY_SOCKET_TIMEOUT = envfloat("XPRA_PROXY_SOCKET_TIMEOUT", "0.1")
+PROXY_WS_TIMEOUT = envfloat("XPRA_PROXY_WS_TIMEOUT", "1.0")
 assert PROXY_SOCKET_TIMEOUT>0, "invalid proxy socket timeout"
+CAN_STOP_PROXY = envbool("XPRA_CAN_STOP_PROXY", False)
 
 
 MAX_CONCURRENT_CONNECTIONS = 200
@@ -52,6 +53,7 @@ class ProxyServer(ServerCore):
         ServerCore.__init__(self)
         self._max_connections = MAX_CONCURRENT_CONNECTIONS
         self._start_sessions = False
+        self.session_type = "proxy"
         self.main_loop = None
         #proxy servers may have to connect to remote servers,
         #or even start them, so allow more time before timing out:
@@ -67,7 +69,6 @@ class ProxyServer(ServerCore):
         self.source_remove = glib.source_remove
         self._socket_timeout = PROXY_SOCKET_TIMEOUT
         self._ws_timeout = PROXY_WS_TIMEOUT
-        self.control_commands["stop"] = ArgsControlCommand("stop", "stops the proxy instance on the given display", self.handle_stop_command, min_args=1, max_args=1)
 
     def init(self, opts):
         log("ProxyServer.init(%s)", opts)
@@ -81,15 +82,22 @@ class ProxyServer(ServerCore):
         get_platform_info()
         self.child_reaper = getChildReaper()
 
-    def init_components(self, opts):
-        pass
+    def init_control_commands(self):
+        ServerCore.init_control_commands(self)
+        self.control_commands["stop"] = ArgsControlCommand("stop", "stops the proxy instance on the given display", self.handle_stop_command, min_args=1, max_args=1)
+
+
+    def make_dbus_server(self):
+        from xpra.server.proxy.proxy_dbus_server import Proxy_DBUS_Server
+        return Proxy_DBUS_Server(self)
+
 
     def init_packet_handlers(self):
         ServerCore.init_packet_handlers(self)
         #add shutdown handler
         self._default_packet_handlers["shutdown-server"] = self._process_proxy_shutdown_server
 
-    def _process_proxy_shutdown_server(self, proto, packet):
+    def _process_proxy_shutdown_server(self, proto, _packet):
         assert proto in self._requests
         self.quit(False)
 
@@ -142,10 +150,6 @@ class ProxyServer(ServerCore):
         self.main_loop.quit()
         log.info("Proxy Server process ended")
 
-    def add_listen_socket(self, socktype, sock):
-        sock.listen(5)
-        glib.io_add_watch(sock, glib.IO_IN, self._new_connection, sock)
-        self.socket_types[sock] = socktype
 
     def verify_connection_accepted(self, protocol):
         #if we start a proxy, the protocol will be closed
@@ -158,10 +162,16 @@ class ProxyServer(ServerCore):
             #already handled in superclass
             return
         self.accept_client(proto, c)
-        if any(c.boolget("%s_request" % x) for x in ("screenshot", "event", "print", "exit")):
+        generic_request = c.strget("request")
+        def is_req(mode):
+            return generic_request==mode or c.boolget("%s_request" % mode)
+        if any(is_req(x) for x in ("screenshot", "event", "print", "exit")):
             self.send_disconnect(proto, "invalid request")
             return
-        if c.boolget("stop_request"):
+        if is_req("stop"):
+            if not CAN_STOP_PROXY:
+                self.send_disconnect(proto, "cannot stop proxy server")
+                return
             self._requests.add(proto)
             #send a hello back and the client should then send its "shutdown-server" packet
             capabilities = self.make_hello()
@@ -174,16 +184,16 @@ class ProxyServer(ServerCore):
                 if not proto._closed:
                     self.send_disconnect(proto, "timeout")
             self.timeout_add(10*1000, force_exit_request_client)
-        else:
-            self.start_proxy(proto, c, auth_caps)
+            return
+        self.proxy_auth(proto, c, auth_caps)
 
-    def start_proxy(self, client_proto, c, auth_caps):
+    def proxy_auth(self, client_proto, c, auth_caps):
         def disconnect(reason, *extras):
             log("disconnect(%s, %s)", reason, extras)
             self.send_disconnect(client_proto, reason, *extras)
 
         #find the target server session:
-        if not client_proto.authenticator:
+        if not client_proto.authenticators:
             log.error("Error: the proxy server requires an authentication mode,")
             try:
                 log.error(" client connection '%s' does not specify one", client_proto._conn.socktype)
@@ -192,32 +202,51 @@ class ProxyServer(ServerCore):
             log.error(" use 'none' to disable authentication")
             disconnect(SESSION_NOT_FOUND, "no sessions found")
             return
-        try:
-            sessions = client_proto.authenticator.get_sessions()
-        except Exception as e:
-            authlog("failed to get the list of sessions", exc_info=True)
-            authlog.error("Error: failed to get the list of sessions using '%s' authenticator", client_proto.authenticator)
-            authlog.error(" %s", e)
-            disconnect(AUTHENTICATION_ERROR, "cannot access sessions")
-            return
-        authlog("start_proxy(%s, {..}, %s) found sessions: %s", client_proto, auth_caps, sessions)
+        sessions = None
+        for authenticator in client_proto.authenticators:
+            try:
+                auth_sessions = authenticator.get_sessions()
+                authlog("proxy_auth %s.get_sessions()=%s", authenticator, auth_sessions)
+                if auth_sessions:
+                    sessions = auth_sessions
+                    break
+            except Exception as e:
+                authlog("failed to get the list of sessions from %s", authenticator, exc_info=True)
+                authlog.error("Error: failed to get the list of sessions using '%s' authenticator", authenticator)
+                authlog.error(" %s", e)
+                disconnect(AUTHENTICATION_ERROR, "cannot access sessions")
+                return
+        authlog("proxy_auth(%s, {..}, %s) found sessions: %s", client_proto, auth_caps, sessions)
         if sessions is None:
             disconnect(SESSION_NOT_FOUND, "no sessions found")
             return
+        self.proxy_session(client_proto, c, auth_caps, sessions)
+
+    def proxy_session(self, client_proto, c, auth_caps, sessions):
+        def disconnect(reason, *extras):
+            log("disconnect(%s, %s)", reason, extras)
+            self.send_disconnect(client_proto, reason, *extras)
         uid, gid, displays, env_options, session_options = sessions
-        if os.name=="posix":
-            if uid==0 or gid==0:
-                log.error("Error: proxy instances should not run as root")
-                log.error(" use a different uid and gid (ie: nobody)")
-                disconnect(AUTHENTICATION_ERROR, "cannot run proxy instances as root")
-                return
+        if POSIX:
+            if getuid()==0:
+                if uid==0 or gid==0:
+                    log.error("Error: proxy instances cannot run as root")
+                    log.error(" use a different uid and gid (ie: nobody)")
+                    disconnect(AUTHENTICATION_ERROR, "cannot run proxy instances as root")
+                    return
+            else:
+                uid = getuid()
+                gid = getgid()
             username = get_username_for_uid(uid)
             groups = get_groups(username)
-            if "xpra" not in groups:
-                log.error("Error: user '%s' (uid=%i) is not in the xpra group", username, uid)
-                log.error(" it belongs to: %s", csv(groups) or None)
-                disconnect(PERMISSION_ERROR, "user missing 'xpra' group membership")
-                return
+            log("username(%i)=%s, groups=%s", uid, username, groups)
+        else:
+            #the auth module recorded the username we authenticate against
+            assert client_proto.authenticators
+            for authenticator in client_proto.authenticators:
+                username = getattr(authenticator, "username", "")
+                if username:
+                    break
         #ensure we don't loop back to the proxy:
         proxy_virtual_display = os.environ.get("DISPLAY")
         if proxy_virtual_display in displays:
@@ -225,48 +254,27 @@ class ProxyServer(ServerCore):
         #remove proxy instance virtual displays:
         displays = [x for x in displays if not x.startswith(":proxy-")]
         #log("unused options: %s, %s", env_options, session_options)
-        opts = make_defaults_struct()
-        display = None
         proc = None
+        socket_path = None
+        display = None
         sns = c.dictget("start-new-session")
-        authlog("start_proxy: displays=%s, start-new-session=%s", displays, bool(sns))
+        authlog("proxy_session: displays=%s, start_sessions=%s, start-new-session=%s", displays, self._start_sessions, sns)
         if len(displays)==0 or sns:
-            if self._start_sessions:
-                #start a new session
-                mode = sns.get("mode", "start")
-                assert mode in ("start", "start-desktop", "shadow"), "invalid start-new-session mode '%s'" % mode
-                sns = typedict(sns)
-                display = sns.get("display")
-                args = []
-                if display:
-                    args = [display]
-                start = sns.strlistget("start")
-                start_child = sns.strlistget("start-child")
-                exit_with_children = sns.boolget("exit-with-children")
-                exit_with_client = sns.boolget("exit-with-client")
-                log("starting new server subprocess: mode=%s, socket-dir=%s, socket-dirs=%s, start=%s, start-child=%s, exit-with-children=%s, exit-with-client=%s, uid=%s, gid=%s",
-                    mode, opts.socket_dir, opts.socket_dirs, start, start_child, exit_with_children, exit_with_client, uid, gid)
-                try:
-                    proc, socket_path = start_server_subprocess(sys.argv[0], args, mode, opts,
-                                                                opts.socket_dir, opts.socket_dirs,
-                                                                start, start_child,
-                                                                exit_with_children, exit_with_client,
-                                                                uid=uid, gid=gid)
-                    display = "socket:%s" % socket_path
-                except Exception as e:
-                    log("start_server_subprocess failed", exc_info=True)
-                    log.error("Error: failed to start server subprocess:")
-                    log.error(" %s", e)
-                    disconnect(SERVER_ERROR, "failed to start a new session")
-                    return
-                if proc:
-                    self.child_reaper.add_process(proc, "server-%s" % display, "xpra start", True, True)
-            else:
+            if not self._start_sessions:
                 disconnect(SESSION_NOT_FOUND, "no displays found")
                 return
+            try:
+                proc, socket_path, display = self.start_new_session(username, uid, gid, sns, displays)
+                log("start_new_session%s=%s", (username, uid, gid, sns, displays), (proc, socket_path, display))
+            except Exception as e:
+                log("start_server_subprocess failed", exc_info=True)
+                log.error("Error: failed to start server subprocess:")
+                log.error(" %s", e)
+                disconnect(SERVER_ERROR, "failed to start a new session")
+                return
         if display is None:
             display = c.strget("display")
-            authlog("start_proxy: proxy-virtual-display=%s (ignored), user specified display=%s, found displays=%s", proxy_virtual_display, display, displays)
+            authlog("proxy_session: proxy-virtual-display=%s (ignored), user specified display=%s, found displays=%s", proxy_virtual_display, display, displays)
             if display==proxy_virtual_display:
                 disconnect(SESSION_NOT_FOUND, "invalid display")
                 return
@@ -280,8 +288,25 @@ class ProxyServer(ServerCore):
                     return
                 display = displays[0]
 
+        connect = c.boolget("connect", True)
+        #ConnectTestXpraClient doesn't want to connect to the real session either:
+        ctr = c.strget("connect_test_request")
+        log("connect=%s, connect_test_request=%s", connect, ctr)
+        if not connect or ctr:
+            log("proxy_session: not connecting to the session")
+            hello = {"display" : display}
+            if socket_path:
+                hello["socket-path"] = socket_path
+            #echo mode if present:
+            mode = sns.get("mode")
+            if mode:
+                hello["mode"] = mode
+            client_proto.send_now(("hello", hello))
+            return
+
         def stop_server_subprocess():
-            if proc:
+            log("stop_server_subprocess() proc=%s", proc)
+            if proc and proc.poll() is None:
                 proc.terminate()
 
         log("start_proxy(%s, {..}, %s) using server display at: %s", client_proto, auth_caps, display)
@@ -292,7 +317,8 @@ class ProxyServer(ServerCore):
             for arg in args:
                 log.warn(" %s", arg)
             raise Exception("parse error on %s: %s" % (display, args))
-        opts.username = client_proto.authenticator.username
+        opts = make_defaults_struct(username=username, uid=uid, gid=gid)
+        opts.username = username
         disp_desc = parse_display_name(parse_error, opts, display)
         if uid or gid:
             disp_desc["uid"] = uid
@@ -303,7 +329,8 @@ class ProxyServer(ServerCore):
         except Exception as e:
             log("cannot connect", exc_info=True)
             log.error("Error: cannot start proxy connection:")
-            log.error(" %s", e)
+            for x in str(e).splitlines():
+                log.error(" %s", x)
             log.error(" connection definition:")
             print_nested_dict(disp_desc, prefix=" ", lchar="*", pad=20, print_fn=log.error)
             disconnect(SESSION_NOT_FOUND, "failed to connect to display")
@@ -323,7 +350,7 @@ class ProxyServer(ServerCore):
         if auth_caps:
             cipher = auth_caps.get("cipher")
             if cipher:
-                encryption_key = self.get_encryption_key(client_proto.authenticator, client_proto.keyfile)
+                encryption_key = self.get_encryption_key(client_proto.authenticators, client_proto.keyfile)
         log("start_proxy(..) client connection=%s", client_conn)
         log("start_proxy(..) client state=%s", client_state)
 
@@ -355,6 +382,114 @@ class ProxyServer(ServerCore):
                 message_queue.put("socket-handover-complete")
         start_thread(do_start_proxy, "start_proxy(%s)" % client_conn)
 
+    def start_new_session(self, username, uid, gid, new_session_dict={}, displays=()):
+        log("start_new_session%s", (username, uid, gid, new_session_dict, displays))
+        sns = typedict(new_session_dict)
+        if WIN32:
+            return self.start_win32_shadow(username, new_session_dict)
+        mode = sns.get("mode", "start")
+        assert mode in ("start", "start-desktop", "shadow"), "invalid start-new-session mode '%s'" % mode
+        display = sns.get("display")
+        if display in displays:
+            raise Exception("display %s is already active!" % display)
+        log("starting new server subprocess: mode=%s, display=%s", mode, display)
+        args = []
+        if display:
+            args = [display]
+        #allow the client to override some options:
+        opts = make_defaults_struct(username=username, uid=uid, gid=gid)
+        for k,v in sns.items():
+            k = bytestostr(k)
+            if k in ("mode", "display"):
+                continue    #those special attributes have been consumed already
+            if k not in PROXY_START_OVERRIDABLE_OPTIONS:
+                log.warn("Warning: ignoring invalid start override")
+                log.warn(" %s=%s", k, v)
+                continue
+            log("start override: %s=%s", k, v)
+            if v is not None:
+                fn = k.replace("-", "_")
+                setattr(opts, fn, v)
+        opts.attach = False
+        opts.start_via_proxy = False
+        env = self.get_proxy_env()
+        cwd = None
+        if uid>0:
+            cwd = get_home_for_uid(uid) or None
+            if not cwd or not os.path.exists(cwd):
+                import tempfile
+                cwd = tempfile.gettempdir()
+        log("starting new server subprocess: options=%s", opts)
+        log("env=%s", env)
+        log("args=%s", args)
+        log("cwd=%s", cwd)
+        proc, socket_path, display = start_server_subprocess(sys.argv[0], args, mode, opts, username, uid, gid, env, cwd)
+        if proc:
+            self.child_reaper.add_process(proc, "server-%s" % (display or socket_path), "xpra %s" % mode, True, True)
+        log("start_new_session(..) pid=%s, socket_path=%s, display=%s, ", proc.pid, socket_path, display)
+        return proc, socket_path, display
+
+    def get_proxy_env(self):
+        ENV_WHITELIST = ["LANG", "HOSTNAME", "PWD", "TERM", "SHELL", "SHLVL", "PATH"]
+        env = dict((k,v) for k,v in os.environ.items() if k in ENV_WHITELIST)
+        #env var to add to environment of subprocess:
+        extra_env_str = os.environ.get("XPRA_PROXY_START_ENV", "")
+        if extra_env_str:
+            extra_env = {}
+            for e in extra_env_str.split(os.path.pathsep):  #ie: "A=1:B=2"
+                parts = e.split("=", 1)
+                if len(parts)==2:
+                    extra_env[parts[0]]= parts[1]
+            log("extra_env(%s)=%s", extra_env_str, extra_env)
+            env.update(extra_env)
+        return env
+
+
+    def start_win32_shadow(self, username, new_session_dict):
+        log("start_win32_shadow%s", (username, new_session_dict))
+        from xpra.platform.paths import get_app_dir
+        from xpra.platform.win32.lsa_logon_lib import logon_msv1_s4u
+        logon_info = logon_msv1_s4u(username)
+        log("logon_msv1_s4u(%s)=%s", username, logon_info)
+        #hwinstaold = set_window_station("winsta0")
+        def exec_command(command):
+            log("exec_command(%s)", command)
+            from xpra.platform.win32.create_process_lib import Popen, CREATIONINFO, CREATION_TYPE_TOKEN, LOGON_WITH_PROFILE, CREATE_NEW_PROCESS_GROUP, STARTUPINFO
+            creation_info = CREATIONINFO()
+            creation_info.dwCreationType = CREATION_TYPE_TOKEN
+            creation_info.dwLogonFlags = LOGON_WITH_PROFILE
+            creation_info.dwCreationFlags = CREATE_NEW_PROCESS_GROUP
+            creation_info.hToken = logon_info.Token
+            log("creation_info=%s", creation_info)
+            startupinfo = STARTUPINFO()
+            startupinfo.lpDesktop = "WinSta0\\Default"
+            startupinfo.lpTitle = "Xpra-Shadow"
+            cwd = get_app_dir()
+            from subprocess import PIPE
+            env = self.get_proxy_env()
+            log("env=%s", env)
+            proc = Popen(command, stdout=PIPE, stderr=PIPE, cwd=cwd, env=env, startupinfo=startupinfo, creationinfo=creation_info)
+            log("Popen(%s)=%s", command, proc)
+            log("poll()=%s", proc.poll())
+            try:
+                log("stdout=%s", proc.stdout.read())
+                log("stderr=%s", proc.stderr.read())
+            except:
+                pass
+            if proc.poll() is not None:
+                return None
+            self.child_reaper.add_process(proc, "server-%s" % username, "xpra shadow", True, True)
+            return proc
+        #whoami = os.path.join(get_app_dir(), "whoami.exe")
+        #exec_command([whoami])
+        port = 10000
+        xpra_command = os.path.join(get_app_dir(), "xpra.exe")
+        command = [xpra_command, "shadow", "--bind-tcp=0.0.0.0:%i" % port, "-d", "win32"]
+        proc = exec_command(command)
+        if not proc:
+            return None, None
+        #exec_command(["C:\\Windows\notepad.exe"])
+        return "tcp/localhost:%i" % port, proc
 
     def reap(self, *args):
         log("reap%s", args)
@@ -368,18 +503,22 @@ class ProxyServer(ServerCore):
             del self.processes[p]
 
 
-    def get_info(self, proto, *args):
+    def get_info(self, proto, *_args):
         info = ServerCore.get_info(self, proto)
         info.setdefault("server", {})["type"] = "Python/GLib/proxy"
         #only show more info if we have authenticated
         #as the user running the proxy server process:
-        pa = proto.authenticator
-        if pa:
-            sessions = pa.get_sessions()
+        if proto and proto.authenticators:
+            sessions = []
+            for authenticator in proto.authenticators:
+                auth_sessions = authenticator.get_sessions()
+                #don't add duplicates:
+                for x in auth_sessions:
+                    if x not in sessions:
+                        sessions.append(x)
             if sessions:
                 uid, gid = sessions[:2]
-                if os.name!="posix" or (uid==os.getuid() and gid==os.getgid()):
-                    info.update(ServerCore.get_info(self, proto))
+                if not POSIX or (uid==os.getuid() and gid==os.getgid()):
                     self.reap()
                     i = 0
                     for p,v in self.processes.items():
